package Apache::AutoID;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
require Exporter;
use CGI::Cookie;
use Carp;
use CGI::Cookie;

BEGIN
{
    if($ENV{MOD_PERL})
    {
        require CGI::Apache;
        require Apache::Constants;
        require Apache::Table;
        Apache::Constants->import(qw(:common REDIRECT M_GET));
    }
}

@ISA = qw(Exporter);
@EXPORT_OK = qw(
    apache_id
    cgi_id
);
$VERSION = '0.01';

sub DEBUG () { 0 };

##########################################################################
use vars qw($has_md5 $max_rand);
BEGIN 
{
    eval { require Digest::MD5; Digest::MD5->import(qw(md5_hex))};
    $has_md5=exists $INC{'Digest/MD5.pm'};
    unless($has_md5)
    {
        require Config;
        Config->import();
        $max_rand=(1<<($Config::Config{randbits}))-1;
    }
}


##########################################################################
use vars qw($id $expire $name $path);


#######################################
sub handler
{
    my($r)=@_;

    DEBUG && $r->warn(__PACKAGE__."::handler");

    # set up a valid request object
    Apache->request($r) unless $r->current_callback eq 'PerlHandler';

    my $id_           = $r->dir_config('AutoID');
    my $want_uri      = $r->dir_config('AutoIDWantURI');
    my $force_uri     = $r->dir_config('AutoIDForceURI');
    my $uri_path      = $r->dir_config('AutoIDURIPath');
    my $want_cookie   = $r->dir_config('AutoIDWantCookie');
    my $force_cookie  = $r->dir_config('AutoIDForceCookie');
    my $cookie_expire = $r->dir_config('AutoIDCookieExpire');
    my $cookie_path   = $r->dir_config('AutoIDCookiePath');

    return apache_id($r, 'id'=>$id_, 
        'want_uri'=>$want_uri, 'force_uri'=>$force_uri, 'uri_path'=>$uri_path, 
        'want_cookie'=>$want_cookie, 'force_cookie'=>$force_cookie,
        'cookie_expire'=>$cookie_expire, 'cookie_path'=>$cookie_path);
}


#######################################
# Set all the default values common to both apache_id and cgi_id
sub _defaults
{
    my($params)=@_;
    $params->{id}||='ID';                       
    $params->{want_uri}=1 unless exists $params->{want_uri};
    $params->{force_uri}=0  unless exists $params->{force_uri};
    $params->{uri_path}||='/';
    $params->{want_cookie}=1  unless exists $params->{want_cookie};
    $params->{force_cookie}=0  unless exists $params->{force_cookie};
    $params->{cookie_expire}||='+1h';
    $params->{cookie_path}||='/';
    $params->{generate_id}=\&generate_id unless exists $params->{generate_id}; 
}

#######################################
sub apache_id
{
    my($r, %params)=@_;

    _defaults(\%params);

    my $id_=$params{id};
    my $phase=$r->current_callback;
    my $method=$r->method();
    DEBUG and $r->warn("$$ $phase : $method");

    my $cookies=$r->header_in('Cookie');
    my %cookies=CGI::Cookie->parse($cookies) if $cookies;
    $cookies{$id_}||='';

    my $hidden='';
    if($phase eq 'PerlHandler')
    {
        my $query=CGI::Apache->new;
        $hidden=$query->param($id_) if $query and defined $query->param($id_);
    } elsif($method eq 'GET')
    {
        my %query=$r->args;
        $hidden=$query{$id_} if exists $query{$id_} and defined $query{$id_};
    } elsif($method eq 'POST')        # turn a GET into a POST
    {                                 # so that we can cooperate with CGI.pm
        my $content = $r->content;
        $r->args($content);
        $r->method('GET');
        $r->method_number(&M_GET);
        $r->headers_in->unset('Content-length');
        my %query=$r->args;
        $hidden=$query{$id_} if exists $query{$id_} and defined $query{$id_};
    } else
    {
        $r->warn("Unable to get parameters for a $method request during $phase");
    }

    DEBUG and $r->warn("$id_ is $hidden") if $hidden;

    my $ret=($phase =~ /Perl(Authen|Authz)Handler/ ? &OK : &DECLINED);

    my $current_id=setup (
        %params,
        uri=>$r->uri(),                     # ways to get at the ID
        cookie=>$cookies{$id_},
        hidden=>$hidden,
        referer=>$r->header_in('Referer'),

        set_cookie=>sub {
            DEBUG && $r->warn("Set cookie $_[0]");
            $r->cgi_header_out('Set-cookie', $_[0])
        },
    
        redirect=>sub {                     # add the ID to the uri
            my($uri, $id)=@_;
            DEBUG && $r->warn("uri=$uri, $id_=$id");
            $uri=~ s(($params{uri_path}(/?)))       
                    ($1$id_$id$2);
            return if $uri eq $r->uri;

            my $to=($r->parsed_uri->scheme||'http').'://';
            $to.=$r->get_server_name;

            my $p=$r->get_server_port;
            $to.=":$p" if $p != 80;
            $to.=$uri;
            my $sub=sub {
                DEBUG && $r->warn("Redirect to $to $$");
                $r->cgi_header_out('Location'=>$to);
                return &REDIRECT;
            };
            if($phase eq 'PerlHandler')     # if in content phase
            {
                $ret=$sub->();              # do redirect now
            } else
            {                               # otherwise, defer it to that phase
                $r->set_handlers('PerlHandler', $sub);
            }
        },

        set_uri=>sub {                      # remove the ID from the uri
            my($uri)=@_;                    # (internaly)
            $r->uri($uri);
        },
    );

    DEBUG && $r->warn("$id_ is $current_id");
    $r->notes($id_=>$current_id);

    $r->register_cleanup(sub {              # kill the ID after this request
        DEBUG && $r->warn('AutoID cleanup');    # just incase
        undef($id);
        undef($expire);
        undef($name);
        undef($path);
        return $ret;
    });
    DEBUG and $r->warn("$$ returning $ret");
    return $ret;

}


#######################################
sub cgi_id
{
    my($query, %params)=@_;
    _defaults(\%params);
    
    my $id_=$params{id};

    my $t_cookie;
    my $uri_now=$ENV{REQUEST_URI};  # for bypassing redirect on rewrite

    return setup {
        uri=>$uri_now,   # ways to get at the ID
        cookie=>$query->cookie($params{id}),
        hidden=>$query->param($params{id}),
        referer=>$ENV{'HTTP_REFERER'},

        set_cookie=>sub {
            DEBUG && warn("Set cookie");    
            $t_cookie=$_[0];
        },
    
        redirect=>sub {                     # add the ID to the uri
            my($uri, $id)=@_;
            DEBUG && warn("uri=$uri, $id_=$id\n");
            $uri=~ s(($params{uri_path}(/?)))            
                    ($1$id_$id$2);

            return if $uri eq $uri_now;     # BYPASS

            ($ENV{SCRIPT_URI}||'')=~m(^(.+)://);
            my $to=($1 || 'http').'://';
            $to.=$ENV{SERVER_NAME};

            my $p=$ENV{SERVER_PORT};
            $to.=":$p" if $p != 80;
            $to.=$uri;

            print $query->redirect(-uri=>$to, -cookies=>$t_cookie);
            exit 0;
        },

        set_uri=>sub {                      # remove the ID from the uri
            my($uri)=@_;                    # (internaly)
            $ENV{REQUEST_URI}=$uri;
            $ENV{SCRIPT_URL}=$uri;
            # TODO also needs to be changed :
            # SCRIPT_URI PATH_TRANSLATED PATH_INFO
        },
    };
}

#######################################
sub setup
{
    my(%params)=@_;
    $id='';
    my $done_cookie=0;
    my $done_redirect=0;
    $params{id}=exists $params{id} ? quotemeta($params{id}) : 'ID';    
    
    $name=$params{'id'};
    $expire=$params{cookie_expire}||'';
    $path=$params{cookie_path}||'/';

    DEBUG && warn __PACKAGE__,  " cookie expires : $expire\n";

    if($params{hidden})             ## Use the ID from a hidden field
    {
        $id=$params{hidden};
        DEBUG && warn __PACKAGE__, " ID from hidden\n";
    }

    if(!$id and $params{uri})       ## or hidden in the URI
    {   
        if($params{uri} =~ m($params{'id'}([-.\w]+)))     # pull it out
        {
            $id=$1;
            DEBUG && warn __PACKAGE__, " ID from URI\n";
            $done_redirect=1;
            if($params{uri} =~ s($params{'id'}$id/?)())   # and set to 'real' uri
            {
                $params{set_uri}->($params{uri})        if $params{set_uri};
            }
        }
    }

    if(!$id and $params{cookie})    ## or from a cookie
    {
        if($params{cookie} =~ m($params{'id'}=([-.\w]+)))     # pull it out
        {
            $id=$1;
            DEBUG && warn __PACKAGE__, " ID from cookie\n";
            # already has cookie, don't send again
            $done_cookie=1   unless $params{force_cookie};     
            $done_redirect=1 unless $params{force_uri};     
        }
    }

    if(!$id and $params{referer})   ## maybe the referer URI had on?
    {
        if($params{referer} =~ m($name([-.\w]+)))  # pull it out
        {
            $id=$1;
            DEBUG && warn __PACKAGE__, " ID from referer\n";
            if($params{redirect})
            {
                $params{redirect}->($params{uri}, $id);
                $done_redirect=1;
            }
        }        
    }
    
    if(!$id and $params{generate_id})   ## no ID! generate one
    {
        $id=$params{generate_id}->();
        DEBUG && warn __PACKAGE__, " generating ID\n";
        if($params{set_cookie})         ## and try to set the cookie
        {
            $params{set_cookie}->(cookie());
            $done_cookie=1;
        }
    }

    return unless $id;                  ## no go...

    ## Set a cookie (if we haven't already)
    $params{set_cookie}->(cookie()) if $params{set_cookie} and 
                                  not $done_cookie and
                                  $params{force_cookie};

    ## put the ID into the URI
    $params{redirect}->($params{uri}, $id) 
                               if $params{redirect} and $params{uri} and
                                  not $done_redirect and 
                                ($params{want_uri} or $params{force_uri});
    return $id;    
}

#######################################
sub hidden_field 
{ 
    return '' unless $id;
    return qq(<INPUT type=hidden name=").$name.qq(" value=").$id.qq(">);
}

#######################################
sub cookie 
{
    return '' unless $id;
    return CGI::Cookie->new(-name=>$name, -value=>$id, 
                            -expires=>$expire, -path=>$path
                           )->as_string;
}

#######################################

sub generate_id
{
    return substr(md5_hex($$, time, rand(), {}, __PACKAGE__), 0, 16)
        if $has_md5;
 
    my $id=sprintf('%X%X%X', time(), $$, rand($max_rand));
    return $id;
}

1;

__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Apache::AutoID - Invisibly (nearly) deal with session IDs

=head1 SYNOPSIS

    use Apache::AutoID;

    # ...

    $id=$r->notes('ID');

and in .htaccess:

  <IfModule mod_perl.c>
    <FilesMatch  ".+\.html$">
      PerlInitHandler Apache::AutoID 

      PerlSetVar AutoIDURIPath /dynamic
      PerlSetVar AutoIDCookiePath /dynamic
      PerlSetVar AutoIDCookieExpire +5h

    </FilesMatch>    
  </IfModule>


=head1 DESCRIPTION

Apache::AutoID is intended to invisibly deal with session IDs.  It is
supposed to work via mod_perl or CGI.  It allows session IDs to be
transmitted via cookies, hidden fields or URI rewrites.  However, URI
rewrites is currently not working.


=head1 FUNCTIONS

=over 2

=item apache_id($r, %parameters)

Wraper around setup() that will interface with an Apache::Request object.


 
Parameters are the following:

    $r = Apache::Request object

    Hash with following keys :    
    id
    cookie_expire
    cookie_path
    force_cookie
    force_uri
    generate_id
    uri_path
    want_cookie
    want_uri

See L<setup()> for their meaning.

Cookies and URI rewriting is handled via $r, and the ID is stashed via
notes().  

Returns Apache handler response code.





=item cgi_id($query, %parameters)

Wraper around setup() that will interface with a CGI object.

Parameters are the following:

    $query = A CGI object

    Hash with following keys :    
    id
    cookie_expire
    cookie_path
    force_cookie
    force_uri
    generate_id
    uri_path
    want_cookie
    want_uri

        See L<setup()> for their meaning.

The only methods called on $query are ->param, ->cookie and ->redirect.  If
you don't use CGI.pm, you may want to create a new class that handles these
methods rather then using the full setup() function.

URI rewriting is handled via $query (ie, the function might not return!).
To set cookies, you must call Apache::AutoID::cookie.

Returns the session ID.




=item cookie()

Returns a cookie string of the ID, suitable for CGI->header :

    print $query->header(-type=>'text/html', 
                         -cookie=>Apache::AutoID::cookie());






=item generate_id()

Default generator of new IDs.  uses Digest::MD5 if it's available, otherwise
simply uses PID, epoch seconds and a random number.

Because I know you won't trust my method of generating session IDs (I know I
wouldn't trust your's), here's the code so you don't have to RTFS:

    $id = substr(Digest::MD5::md5_hex($$, time, rand(), 
                                        {}, __PACKAGE__), 0, 16);

I believe that 64 bits will be sufficient to avoid duplicating an ID. 
While 128 bits would be much much much safer, it would also be unweildly for
URI rewriting.

Digest::MD5 is part of Systemics' excellent cryptography bundle, Cryptix. 
ftp://ftp.funet.fi/pub/crypt/cryptography/libs/perl/

If Digest::MD5 is not installed, the following is used to generate session
IDs:

    use Config;
    $max_rand=(1<<($Config{randbits}))-1;
    $id=sprintf('%X%X%X', time(), $$, rand($max_rand));





=item handler($r)

This is the Apache handler.  It should be called from PerlInitHandler. 
However, see BUGS below.

It can use the following PerlVars:
    
    SetPerlVar          setup() equiv
    AutoID              id
    AutoIDWantURI       want_uri 
    AutoIDForceURI      force_uri 
    AutoIDURIPath       uri_path 
    AutoIDWantCookie    want_cookie 
    AutoIDForceCookie   force_cookie 
    AutoIDCookieExpire  cookie_expire
    AutoIDCookiePath    cookie_path 





=item hidden_field()

Return an hidden HTML input field for use in a form.

    print qq(<FORM method=POST action="$cgi">\n),
            Apache::AutoID::hidden_field();





=item setup(...)

Do the dirty work of figuring out how to get at the ID.  Parameters as
follows :

    id [ID]

This is the name of the ID.  This is used for cookie name, notes() field uri
rewriting, and HTTP paramter.  Through out this document, ID means the value
of this parameter.

    hidden 
    cookie
    uri
    generate_id     [\&Apache::AutoID::generate_id]

These are the source of the ID.   If hidden is set, it's value is taken for
that of the ID.  If not, URI is checked for m/ID([.\w]+)/ and $1 is the ID. 
If it's there, set_uri() is called, if present.  If we still don't have an ID, the
cookie is looked at.  Then referer and if all those fail, generate_id() is
called.

    cookie_expire   [+1h]
    cookie_path     [/]
    force_cookie    [0]
    set_cookie 
    want_cookie     [1]

If want_cookie is true a cookie will be sent to the browser when a new ID is
generated.  if force_cookie is true, a cookie is sent no matter what. This
will be very annoying for those that have their browser set to prompt the
user to accept/deny each cookie.

To set the cookie, set_cookie() is called with one param : a cookie that has
the path of cookie_path that will expire at cookie_expire.  Cookie_expire
can be the for +5d (5 days) or +10h (10 hours).

    force_uri       [0]
    redirect 
    referer
    set_uri 
    uri_path        [/]
    want_uri        [1]

URI rewrite code is currently buggy.  What I want to do is the following:

If no ID is present in the URI, do a redirect to new ID xxxx :

    /foo/bar.html => /foo/IDxxxx/bar.html (note, uri_path = /foo)

force_uri would require this to be done, no matter what.  If an absolute
link is followed, we check the referer to to see if the ID was there. 
set_uri would be used to set various bits of the environment and so on to
reflect the URI without the ID "hidden" in it.  This becomes complicated in
the case of CGI (ie PATH_TRANSLATD, SCRIPT_PATH SCRIPT_URI, REQUEST_URI) and
so on and I haven't taken the time to work it all out.  If you fix this,
please tell me!


=back




=head1 BUGS

Apache::Request::read is not usable when called during the Init phase.  This
means that CGI.pm will not be able to read and parse POSTed data.  To work
around this, a POST request is turned into a GET request unless invoked
during the content phase.  This means that Apache::AutoID will cooperate
with CGI.pm and CGI::Lite and most other modules, but you might find it
suprising at some point.  

Note that the request will still be logged as a POST request in access_log. 
So sensitive information in HTML fields won't appear there.




=head1 AUTHOR

Philip Gwyn <fil@pied.nu>

=head1 SEE ALSO

L<Apache::AutoSession>, L<Digest::MD5>, perl(1).

=cut
